@@ -1,7 +1,7 @@
 use crate::{
     marker::*,
-    secp256k1::{PublicKey, SecretKey},
-    Point, Scalar,
+    secp256k1::{PublicKey, SecretKey, XOnlyPublicKey},
+    Point, Scalar, XOnly,
 };
 
 impl From<Scalar> for SecretKey {
@@ -31,6 +31,30 @@ impl<T: Normalized> From<Point<T>> for PublicKey {
     }
 }
 
+impl From<XOnlyPublicKey> for XOnly {
+    fn from(pk: XOnlyPublicKey) -> Self {
+        XOnly::from_bytes(pk.serialize()).unwrap()
+    }
+}
+
+impl From<XOnly> for XOnlyPublicKey {
+    fn from(xonly: XOnly) -> Self {
+        XOnlyPublicKey::from_slice(xonly.as_bytes()).unwrap()
+    }
+}
+
+impl From<Point<EvenY>> for XOnlyPublicKey {
+    fn from(point: Point<EvenY>) -> Self {
+        point.to_xonly().into()
+    }
+}
+
+impl From<XOnlyPublicKey> for Point<EvenY> {
+    fn from(pk: XOnlyPublicKey) -> Self {
+        XOnly::from(pk).to_point()
+    }
+}
+
 #[cfg(test)]
 mod test {
     use super::*;
