@@ -1,6 +1,5 @@
 //! Scalar arithmetic (integers mod the secp256k1 group order)
 use crate::{backend, hash::HashInto, marker::*, op};
-use backend::BackendScalar;
 use core::marker::PhantomData;
 use digest::{generic_array::typenum::U32, Digest};
 use rand_core::{CryptoRng, RngCore};
@@ -50,6 +49,8 @@ use rand_core::{CryptoRng, RngCore};
 #[derive(Clone, Eq)]
 pub struct Scalar<S = Secret, Z = NonZero>(pub(crate) backend::Scalar, PhantomData<(Z, S)>);
 
+impl<Z: Clone> Copy for Scalar<Public, Z> {}
+
 impl<Z> core::hash::Hash for Scalar<Public, Z> {
     fn hash<H: core::hash::Hasher>(&self, state: &mut H) {
         self.to_bytes().hash(state)
@@ -59,7 +60,7 @@ impl<Z> core::hash::Hash for Scalar<Public, Z> {
 impl<Z, S> Scalar<S, Z> {
     /// Serializes the scalar to its 32-byte big-endian representation
     pub fn to_bytes(&self) -> [u8; 32] {
-        backend::Scalar::to_bytes(&self.0)
+        backend::BackendScalar::to_bytes(&self.0)
     }
 
     /// Negates the scalar in-place if `cond` is true.
@@ -147,7 +148,7 @@ impl Scalar<Secret, NonZero> {
     ///
     /// [`NonZeroU32`]: core::num::NonZeroU32
     pub fn from_non_zero_u32(int: core::num::NonZeroU32) -> Self {
-        Self::from_inner(backend::Scalar::from_u32(int.get()))
+        Self::from_inner(backend::BackendScalar::from_u32(int.get()))
     }
 
     /// Returns the integer `1` as a `Scalar<Secret, NonZero>`.
@@ -157,7 +158,7 @@ impl Scalar<Secret, NonZero> {
 
     /// Returns the integer -1 (modulo the curve order) as a `Scalar<Secret, NonZero>`.
     pub fn minus_one() -> Self {
-        Self::from_inner(backend::Scalar::minus_one())
+        Self::from_inner(backend::BackendScalar::minus_one())
     }
 }
 
@@ -176,7 +177,7 @@ impl Scalar<Secret, Zero> {
     /// assert_eq!(scalar_overflowed, Scalar::one())
     /// ```
     pub fn from_bytes_mod_order(bytes: [u8; 32]) -> Self {
-        Self::from_inner(backend::Scalar::from_bytes_mod_order(bytes))
+        Self::from_inner(backend::BackendScalar::from_bytes_mod_order(bytes))
     }
 
     /// Exactly like [`from_bytes_mod_order`] except
@@ -214,7 +215,7 @@ impl Scalar<Secret, Zero> {
     /// assert!(Scalar::from_bytes([255u8; 32]).is_none());
     /// ```
     pub fn from_bytes(bytes: [u8; 32]) -> Option<Self> {
-        backend::Scalar::from_bytes(bytes).map(Self::from_inner)
+        backend::BackendScalar::from_bytes(bytes).map(Self::from_inner)
     }
 
     /// Creates a scalar from 32 big-endian encoded bytes in a slice. If the
@@ -238,7 +239,30 @@ impl Scalar<Secret, Zero> {
     /// assert_eq!(s!(zero * x), zero);
     /// assert_eq!(s!(x + zero), x);
     pub fn zero() -> Self {
-        Self::from_inner(backend::Scalar::zero())
+        Self::from_inner(backend::BackendScalar::zero())
+    }
+}
+
+impl<S> Scalar<S, Zero> {
+    /// Converts a scalar marked with `Zero` to one that is marked `NonZero`.
+    /// You must provide a justification for this as the `reason`.
+    /// **If you're wrong the method will panic with the reason**.
+    ///
+    /// This is shorthand for:
+    ///
+    /// ```ignore
+    /// use secp256kfun::marker::{Mark, NonZero};
+    /// scalar.mark::<NonZero>().expect(reason);
+    /// ```
+    ///
+    /// # Example
+    ///
+    /// ```
+    /// use secp256kfun::{s, Scalar};
+    /// let two = s!(1 + 1).expect_nonzero("one plus one is not zero");
+    /// ```
+    pub fn expect_nonzero(self, reason: &str) -> Scalar<S, NonZero> {
+        self.mark::<NonZero>().expect(reason)
     }
 }
 
@@ -250,7 +274,7 @@ impl<Z1, Z2, S1, S2> PartialEq<Scalar<S2, Z2>> for Scalar<S1, Z1> {
 
 impl From<u32> for Scalar<Secret, Zero> {
     fn from(int: u32) -> Self {
-        Self::from_inner(backend::Scalar::from_u32(int))
+        Self::from_inner(backend::BackendScalar::from_u32(int))
     }
 }
 
@@ -292,8 +316,8 @@ impl<S, Z> core::ops::Neg for &Scalar<S, Z> {
     }
 }
 
-impl HashInto for Scalar {
-    fn hash_into(&self, hash: &mut impl digest::Digest) {
+impl<S, Z> HashInto for Scalar<S, Z> {
+    fn hash_into(self, hash: &mut impl digest::Digest) {
         hash.update(&self.to_bytes())
     }
 }
@@ -320,6 +344,7 @@ where
 mod test {
     use super::*;
     use crate::{hex, op, s};
+    use proptest::prelude::*;
     #[cfg(target_arch = "wasm32")]
     use wasm_bindgen_test::wasm_bindgen_test as test;
 
@@ -339,17 +364,49 @@ mod test {
         assert_ne!(scalar_1, scalar_2);
     }
 
-    #[test]
-    fn invert() {
-        let x = Scalar::random(&mut rand::thread_rng());
-        assert!(s!(x * { x.invert() }) == Scalar::from(1));
+    proptest! {
+        #[test]
+        fn invert(x in any::<Scalar>(), y in any::<Scalar<Public>>()) {
+            assert_eq!(s!(x * { x.invert() }), s!(1));
+            assert_eq!(s!(y * { y.invert() }), s!(1));
+        }
+
+        #[test]
+        fn sub(a in any::<Scalar>(),
+               b in any::<Scalar<Public>>(),
+               c in any::<Scalar<Public,Zero>>(),
+               d in any::<Scalar<Secret,Zero>>(),
+        ) {
+            assert_eq!(s!(a - a), s!(0));
+            assert_eq!(s!(b - b), s!(0));
+            assert_eq!(s!(c - c), s!(0));
+            assert_eq!(s!(d - d), s!(0));
+            assert_eq!(s!(a - a), s!(-a + a));
+            assert_eq!(s!(a - b), s!(-b + a));
+            assert_eq!(s!(a - c), s!(-c + a));
+            assert_eq!(s!(a - d), s!(-d + a));
+
+            if a != b {
+                assert_ne!(s!(a - b), s!(b - a));
+            }
+
+            if c != d {
+                assert_ne!(s!(c - d), s!(d - c));
+            }
+        }
+
+
     }
 
     #[test]
-    fn neg() {
-        let x = Scalar::random(&mut rand::thread_rng());
-        assert_eq!(s!(x - x), Scalar::zero());
-        assert_eq!(-Scalar::zero(), Scalar::zero())
+    fn scalar_subtraction_is_not_commutative() {
+        let two = s!(2);
+        let three = s!(3);
+        let minus_1 = s!(-1);
+        let one = s!(1);
+
+        assert_eq!(s!(two - three), minus_1);
+        assert_eq!(s!(three - two), one);
     }
 
     #[test]
@@ -402,14 +459,9 @@ mod test {
     }
 
     #[test]
-    fn scalar_subtraction_is_not_commutative() {
-        let two = Scalar::from(2);
-        let three = Scalar::from(3);
-        let minus_1 = Scalar::minus_one();
-        let one = Scalar::from(1);
-
+    fn minus_one() {
         assert_eq!(
-            minus_1,
+            Scalar::minus_one(),
             Scalar::from_bytes_mod_order(
                 hex::decode_array(
                     "FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFEBAAEDCE6AF48A03BBFD25E8CD0364140"
@@ -417,8 +469,6 @@ mod test {
                 .unwrap()
             )
         );
-        assert_eq!(s!(two - three), minus_1);
-        assert_eq!(s!(three - two), one);
     }
 
     #[test]
